// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file ParallelProblem.h */

#ifndef AMDIS_PARALLELPROBLEM_H
#define AMDIS_PARALLELPROBLEM_H

#include "ProblemTimeInterface.h"
#include "ProblemIterationInterface.h"
#include "AdaptInfo.h"
#include "mpi.h"
#include <map>
#include <vector>
#include <set>

namespace AMDiS {

  // =========================================================================
  // ===== class ParallelProblemInterface ====================================
  // =========================================================================
  
  /** \brief
   * Interface for parallel problems
   */
  class ParallelProblemInterface
  {
  public:
    virtual void initParallelization(AdaptInfo *adaptInfo) = 0;
    virtual void exitParallelization(AdaptInfo *adaptInfo) = 0;
  };

  // =========================================================================
  // ===== class ParallelProblemBase =========================================
  // =========================================================================

  /** \brief
   * Schablonen Klasse
   */
  class ParallelProblemBase : public ParallelProblemInterface,
			      public ProblemIterationInterface,
			      public ProblemTimeInterface
  {
  public:
    ParallelProblemBase(ProblemIterationInterface *iterationIF,
			ProblemTimeInterface *timeIF);

    virtual ~ParallelProblemBase() {};

    virtual bool doPartitioning(AdaptInfo *adaptInfo, double localWeightSum) = 0;
    virtual bool doBuildGlobalSolution(AdaptInfo *adaptInfo) = 0;
    virtual double setElemWeights(AdaptInfo *adaptInfo) = 0;
    virtual void partitionMesh(AdaptInfo *adaptInfo) = 0;
    virtual void refineOverlap(AdaptInfo *adaptInfo) = 0;    
    virtual void globalRefineOutOfPartition(AdaptInfo *adaptInfo) = 0;    
    virtual void createOverlap(AdaptInfo *adaptInfo) = 0;
    virtual void exchangeDOFVectors(AdaptInfo *adaptInfo) = 0;
    virtual void coarsenOutOfPartition(AdaptInfo *adaptInfo) = 0;
    virtual void synchronizeMeshes(AdaptInfo *adaptInfo) = 0;
    virtual void exchangeRankSolutions(AdaptInfo *adaptInfo) = 0;
    virtual void buildGlobalSolution(AdaptInfo *adaptInfo) = 0;

    virtual void exitParallelization(AdaptInfo *adaptInfo)
    {
      if (!timeIF_) 
	closeTimestep(adaptInfo);
    };

    virtual void setTime(AdaptInfo *adaptInfo) 
    {
      if (timeIF_) 
	timeIF_->setTime(adaptInfo);
    };

    virtual void initTimestep(AdaptInfo *adaptInfo)
    {
      if (timeIF_) 
	timeIF_->initTimestep(adaptInfo);
    };

    virtual void closeTimestep(AdaptInfo *adaptInfo)
    {
      if (mpiSize_ > 1 && doBuildGlobalSolution(adaptInfo)) {
	synchronizeMeshes(adaptInfo);
 	exchangeRankSolutions(adaptInfo);
 	buildGlobalSolution(adaptInfo);
      }
      if (timeIF_) 
	timeIF_->closeTimestep(adaptInfo);
    };

    virtual void solveInitialProblem(AdaptInfo *adaptInfo)
    {
      if (timeIF_)
	timeIF_->solveInitialProblem(adaptInfo);
    };
  
    virtual void transferInitialSolution(AdaptInfo *adaptInfo)
    {
      if (timeIF_) 
	timeIF_->transferInitialSolution(adaptInfo);
    };  


    virtual void beginIteration(AdaptInfo *adaptInfo)
    {
      iterationIF_->beginIteration(adaptInfo);
    };

    virtual Flag oneIteration(AdaptInfo *adaptInfo, Flag toDo = FULL_ITERATION)
    {
      Flag flag;

      if(mpiSize_ > 1 && toDo.isSet(MARK | ADAPT)) {
 	flag = iterationIF_->oneIteration(adaptInfo, MARK | ADAPT);

	double localWeightSum = setElemWeights(adaptInfo);
	if(doPartitioning(adaptInfo, localWeightSum)) {

	  clock_t partitioningStart = clock();

	  synchronizeMeshes(adaptInfo);
	  partitionMesh(adaptInfo);
	  refineOverlap(adaptInfo);
	  createOverlap(adaptInfo);
	  synchronizeMeshes(adaptInfo);
	  exchangeDOFVectors(adaptInfo);
	  coarsenOutOfPartition(adaptInfo);

	  clock_t partitioningEnd = clock();
	  partitioningTime = TIME_USED(partitioningStart, 
				       partitioningEnd);
	  computationStart = partitioningEnd;
	}

	flag |= iterationIF_->oneIteration(adaptInfo, toDo & ~(MARK | ADAPT));
      } else {
	flag = iterationIF_->oneIteration(adaptInfo, toDo);
      }

      // synchronize adaption flag
      unsigned long *flagBuffer = GET_MEMORY(unsigned long, mpiSize_);

      unsigned long localFlag = flag.getFlags();

      MPI::COMM_WORLD.Allgather(&localFlag, 1, MPI_UNSIGNED_LONG,
				flagBuffer, 1, MPI_UNSIGNED_LONG);
      int i;
      for(i = 0; i < mpiSize_; i++) {
	flag.setFlag(flagBuffer[i]);
      }
      FREE_MEMORY(flagBuffer, unsigned long, mpiSize_);
      return flag;
    };

    virtual void endIteration(AdaptInfo *adaptInfo) 
    {
      iterationIF_->endIteration(adaptInfo);
    };

  protected:
    int mpiRank_;
    int mpiSize_;
    ProblemIterationInterface *iterationIF_;
    ProblemTimeInterface *timeIF_;
    clock_t computationStart;
    double partitioningTime;
  };

  // =========================================================================
  // ===== class ParallelProblem =============================================
  // =========================================================================

  class ParMetisPartitioner;
  class RefinementManager;
  class CoarseningManager;
  class ElInfo;
  class MeshStructure;

  class ParallelProblem : public ParallelProblemBase
  {
  public:
    ParallelProblem(const std::string& name,
		    ProblemIterationInterface *iterationIF,
		    ProblemTimeInterface *timeIF,
		    std::vector<DOFVector<double>*> vectors,
		    Mesh *mesh,
		    RefinementManager *rm,
		    CoarseningManager *cm);

    virtual ~ParallelProblem();

    virtual bool doPartitioning(AdaptInfo *adaptInfo, double localWeightSum);

    virtual bool doBuildGlobalSolution(AdaptInfo *adaptInfo);

    virtual void partitionMesh(AdaptInfo *adaptInfo);
    virtual void refineOverlap(AdaptInfo *adaptInfo);
    virtual void createOverlap(AdaptInfo *adaptInfo);    
    virtual void globalRefineOutOfPartition(AdaptInfo *adaptInfo);
    virtual void coarsenOutOfPartition(AdaptInfo *adaptInfo);
    virtual void synchronizeMeshes(AdaptInfo *adaptInfo);
    virtual double setElemWeights(AdaptInfo *adaptInfo);

    virtual void globalRefinements();

    void exchangeRankSolutions(AdaptInfo *adaptInfo,
			       std::vector<DOFVector<double>*> rankSolutions);

    void exchangeDOFVector(AdaptInfo *adaptInfo,
			   DOFVector<double> *vec);

    void buildGlobalSolution(AdaptInfo *adaptInfo,
			     std::vector<DOFVector<double>*> rankSolutions,
			     DOFVector<double> *globalSolution);

    void createOverlap(int level, int overlap, bool openOverlap,
		       std::map<Element*, int> &overlapDistance);

    void fillVertexPartitions(int level, int overlap, bool openOverlap,
			      std::map<Element*, int> &overlapDistance);

    void setRepartitionSteps(int steps) { repartitionSteps_ = steps; };

    void puEveryTimestep(bool pu) { puEveryTimestep_ = pu; };

    void addDOFVector(DOFVector<double> *vec) 
    {
      dofVectors_.push_back(vec);
    };

    void exchangeMeshStructureCodes(MeshStructure *structures);

    static bool writeElement(ElInfo *elInfo);

    //     static void writeRankMacroAndValues(DOFVector<double> *vec, 
    // 					const char *name = NULL, 
    // 					double time = 0.0);


    virtual void serialize(std::ostream&) {};

    virtual void deserialize(std::istream&) {};

  protected:
    double errors2map(std::map<int, double> &errMap, int comp, bool add);

  protected:
    /** \brief
     *
     */
    std::string name_;

    /** \brief
     *
     */
    Mesh *mesh_;

    /** \brief
     *
     */
    RefinementManager *refinementManager_;

    /** \brief
     *
     */
    CoarseningManager *coarseningManager_;

    /** \brief
     * Pointer to the paritioner which is used to devide a mesh into partitions.
     */
    ParMetisPartitioner *partitioner_;

    /** \brief
     * Stores to every element index the number of the partition it corresponds to.
     */
    std::map<int, int> partitionVec_;

    /** \brief
     * Stores an old partitioning of elements. To every element index the number of
     * the parition it corresponds to is stored.
     */
    std::map<int, int> oldPartitionVec_;    

    /** \brief
     *
     */
    std::map<int, double> elemWeights_;

    /** \brief
     *
     */
    std::map<Element*, std::set<int> > elementPartitions_;

    /** \brief
     * Stores to every DOF the set of partitions it corresponds to.
     */
    std::map<DegreeOfFreedom, std::set<int> > vertexPartitions_;

    /** \brief
     *
     */
    int repartitionSteps_;

    /** \brief
     *
     */
    bool puEveryTimestep_;

    /** \brief
     *
     */
    std::vector<DOFVector<double>*> dofVectors_;

    /** \brief
     *
     */
    double upperPartThreshold_;

    /** \brief
     *
     */
    double lowerPartThreshold_;

    /** \brief
     *
     */
    int globalCoarseGridLevel_;

    /** \brief
     *
     */
    int localCoarseGridLevel_;

    /** \brief
     *
     */
    int globalRefinements_;

    /** \brief
     *
     */
    std::map<Element*, int> overlapDistance_;

    /** \brief
     *
     */
    int adaptiveThresholds_;

    /** \brief
     *
     */
    double thresholdIncFactor_;

    /** \brief
     *
     */
    double thresholdDecFactor_;

    /** \brief
     *
     */
    double repartTimeFactor_;

    /** \brief
     *
     */
    double minUpperTH_;

    /** \brief
     *
     */
    double maxLowerTH_;
  };

  // =========================================================================
  // ===== class ParallelProblemScal =========================================
  // =========================================================================
  
  class FiniteElemSpace;
  class ProblemScal;
  class ProblemInstatScal;
  class Estimator;
  class Marker;

  class ParallelProblemScal : public ParallelProblem
  {
  public:
    ParallelProblemScal(const std::string& name,
			ProblemScal *problem,
			ProblemInstatScal *problemInstat,
			std::vector<DOFVector<double>*> vectors);

    virtual ~ParallelProblemScal();

    virtual void initParallelization(AdaptInfo *adaptInfo);

    virtual void exitParallelization(AdaptInfo *adaptInfo);

    virtual void exchangeDOFVectors(AdaptInfo *adaptInfo);

    virtual void exchangeRankSolutions(AdaptInfo *adaptInfo);

    virtual void buildGlobalSolution(AdaptInfo *adaptInfo);

    virtual void coarsenOutOfPartition(AdaptInfo *adaptInfo);

    virtual int getNumProblems();

    virtual ProblemStatBase *getProblem(int number = 0);

    virtual ProblemStatBase *getProblem(const std::string& name);

    void setEstimator(Estimator *est) { 
      usersEstimator_ = est; 
    };

    void setMarker(Marker *marker) { 
      usersMarker_ = marker; 
    };

    inline virtual const ::std::string& getName() { 
      return name_; 
    };

  protected:
    ProblemScal *problem_;

    ProblemInstatScal *problemInstat_;

    Estimator *oldEstimator_;

    Marker *oldMarker_;

    std::vector<DOFVector<double>*> rankSolution_;

    Estimator *usersEstimator_;

    Marker *usersMarker_;
  };

  // =========================================================================
  // ===== class ParallelProblemVec ==========================================
  // =========================================================================
  
  class ProblemVec;
  class ProblemInstatVec;
  class SystemVector;

  class ParallelProblemVec : public ParallelProblem
  {
  public:
    ParallelProblemVec(const std::string& name,
		       ProblemVec *problem,
		       ProblemInstatVec *problemInstat,
		       std::vector<DOFVector<double>*> vectors);

    virtual ~ParallelProblemVec();

    virtual void initParallelization(AdaptInfo *adaptInfo);

    virtual void exitParallelization(AdaptInfo *adaptInfo);

    virtual void exchangeDOFVectors(AdaptInfo *adaptInfo);

    virtual void exchangeRankSolutions(AdaptInfo *adaptInfo);

    virtual void buildGlobalSolution(AdaptInfo *adaptInfo);

    virtual void coarsenOutOfPartition(AdaptInfo *adaptInfo);

    virtual int getNumProblems();

    virtual ProblemStatBase *getProblem(int number = 0);

    virtual ProblemStatBase *getProblem(const std::string& name);

    void setEstimator(std::vector<Estimator*> est) { 
      usersEstimator_ = est; 
    };

    void setMarker(std::vector<Marker*> marker) {
      usersMarker_ = marker; 
    };

    inline virtual const ::std::string& getName() { 
      return name_; 
    };

  protected:
    ProblemVec *problem_;

    ProblemInstatVec *problemInstat_;

    std::vector<Estimator*> oldEstimator_;

    std::vector<Marker*> oldMarker_;

    std::vector<SystemVector*> rankSolution_;

    std::vector<Estimator*> usersEstimator_;

    std::vector<Marker*> usersMarker_;

    int numComponents_;
  };

}

#endif
